import type { Color, LinearRGB, Oklab, Oklch, sRGB, XYZ_D65 } from './colorspace.js';
export interface ColorOutput {
    /** `#000000` */
    hex: string;
    /** `rgb(0, 0, 0)` */
    rgb: string;
    /** `rgba(0, 0, 0, 1)` */
    rgba: string;
    /** 0x000000 */
    hexVal: number;
    /** [R, G, B, alpha] */
    rgbVal: sRGB;
    /** [R, G, B, alpha] */
    rgbaVal: sRGB;
    /** [R, G, B, alpha] */
    linearRGB: LinearRGB;
    /** `color(luv 0 0 0/1)` */
    /** [L, u, v, alpha] */
    /** `color(display-p3 0 0 0/1)` */
    p3: string;
    /** [R, G, B, alpha] */
    p3Val: sRGB;
    /** `color(oklab 0 0 0/1)` */
    oklab: string;
    /** [L, a, b, alpha] */
    oklabVal: Oklab;
    /** `color(oklch 0 0 0/1)` */
    oklch: string;
    /** [L, C, h, alpha] */
    oklchVal: Oklch;
    /** `color(xyz 0 0 0/1)` (2•, D65 whitepoint) */
    xyz: string;
    /** [X, Y, Z, alpha] (2•, D65 whitespace) */
    xyzVal: XYZ_D65;
    toString(): string;
}
/**
 * Parse any valid CSS color color string and convert to:
 * - hex:       '#ff0000'
 * - hexVal:    0xff0000
 * - rgb:       'rgb(255, 0, 0)'
 * - rgbVal:    [1, 0, 0, 1]
 * - rgba:      'rgba(255, 0, 0, 1)'
 * - rgbaVal:   [1, 0, 0, 1]   // alias of rgbVal
 * - linearRGB: [1, 0, 0, 0]   // https://en.wikipedia.org/wiki/Gamma_correction
 * - hslVal:    [0, 1, 1, 1]   // 0% = 0, 100% = 1
 * - p3:        'color(display-p3 1 0 0)'
 * - p3Val:     '[1, 0, 0, 0]'
 */
export declare function from(rawColor: Color): ColorOutput;
/**
 * hex num to sRGB (doesn’t support alpha as 0x000000 == 0x00000000)
 * V8 handles number ops ~ 2x faster than parseInt(hex, 16) with string manipulations
 */
export declare function hexNumTosRGB(hex: number): sRGB;
/** Convert any number of user inputs into RGBA array */
export declare function parse(rawColor: Color): sRGB;
/**
 * Lightness
 * Shortcut of "L” from oklab
 */
export declare function lightness(color: Color): number;
